"""
Tests for the G90Sensor class.
"""
from __future__ import annotations
import asyncio
from unittest.mock import MagicMock
from contextlib import nullcontext, AbstractContextManager
import pytest
from pyg90alarm import (
    G90Alarm, G90Error, G90EntityRegistrationError,
)
from pyg90alarm.entities.sensor import (
    G90Sensor, G90SensorUserFlags, G90SensorAlertModes,
)
from .device_mock import DeviceMock


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[3,1,3],["Remote 1",10,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 2",11,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Cord 1",12,0,126,1,0,32,0,5,16,1,0,""]'
    b']]IEND\0',
])
async def test_multiple_sensors_shorter_than_page(
    mock_device: DeviceMock
) -> None:
    """
    Tests for retrieving multiple sensors from the panel, while the number of
    those is shorter than a single page.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    sensors = await g90.sensors

    assert await mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,10]]]IEND\0',
    ]
    assert len(sensors) == 3
    assert isinstance(sensors, list)
    assert isinstance(sensors[0], G90Sensor)
    assert sensors[0].name == 'Remote 1'
    assert sensors[0].index == 10
    assert sensors[0].is_wireless is True
    assert isinstance(sensors[1], G90Sensor)
    assert sensors[1].name == 'Remote 2'
    assert sensors[1].index == 11
    assert sensors[1].is_wireless is True
    assert isinstance(sensors[2], G90Sensor)
    assert sensors[2].name == 'Cord 1'
    assert sensors[2].index == 12
    assert sensors[2].is_wireless is False


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[11,1,10],'
    b'["Remote 1",10,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 2",11,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 3",12,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 4",13,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 5",14,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 6",15,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 7",16,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 8",17,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 9",18,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 10",19,0,10,1,0,32,0,0,16,1,0,""]'
    b']]IEND\0',
    b'ISTART[102,'
    b'[[11,11,1],'
    b'["Remote 11",20,0,10,1,0,32,0,0,16,1,0,""]'
    b']]IEND\0',
])
async def test_multiple_sensors_longer_than_page(
    mock_device: DeviceMock
) -> None:
    """
    Tests for retrieving multiple sensors from the panel, while the number of
    those is longer than a single page.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    sensors = await g90.sensors

    assert await mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,10]]]IEND\0',
        b'ISTART[102,102,[102,[11,11]]]IEND\0',
    ]
    assert len(sensors) == 11
    assert isinstance(sensors, list)
    assert isinstance(sensors[0], G90Sensor)
    assert sensors[0].name == 'Remote 1'
    assert sensors[0].index == 10


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[1,1,1],["Remote",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
])
async def test_single_sensor(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving single sensor from the panel.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    sensors = await g90.sensors

    assert await mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,10]]]IEND\0',
    ]
    assert len(sensors) == 1
    assert isinstance(sensors, list)
    assert isinstance(sensors[0], G90Sensor)
    assert sensors[0].name == 'Remote'
    assert sensors[0].index == 10
    assert isinstance(sensors[0]._asdict(), dict)


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[2,1,2],'
    b'["Cord 1",11,0,126,1,0,63,0,5,16,1,0,""],'
    b'["Cord 2",10,0,126,1,0,63,0,5,16,1,0,""]'
    b']]IEND\0',
    b'ISTART[102,'
    b'[[2,2,1],'
    b'["Cord 2",10,0,126,1,0,63,0,5,16,1,0,""]'
    b']]IEND\0',
    b"ISTARTIEND\0",
])
@pytest.mark.parametrize(
    'flag,value_before,value_after,expected_data', [
        pytest.param(
            G90SensorUserFlags.ENABLED, True, False, [
                b'ISTART[102,102,[102,[1,10]]]IEND\0',
                b'ISTART[102,102,[102,[2,2]]]IEND\0',
                b'ISTART[103,103,[103,'
                b'["Cord 2",10,0,126,1,0,62,0,5,16,1,0,0,"00"]'
                b']]IEND\0',
            ],
            id='enabled',
        ),
        pytest.param(
            G90SensorUserFlags.ENABLED, True, True, [
                b'ISTART[102,102,[102,[1,10]]]IEND\0',
            ],
            id='enabled-not-changed',
        ),
        pytest.param(
            G90SensorUserFlags.ARM_DELAY, True, False, [
                b'ISTART[102,102,[102,[1,10]]]IEND\0',
                b'ISTART[102,102,[102,[2,2]]]IEND\0',
                b'ISTART[103,103,[103,'
                b'["Cord 2",10,0,126,1,0,61,0,5,16,1,0,0,"00"]'
                b']]IEND\0',
            ],
            id='arm_delay',
        ),
        pytest.param(
            G90SensorUserFlags.DETECT_DOOR, True, False, [
                b'ISTART[102,102,[102,[1,10]]]IEND\0',
                b'ISTART[102,102,[102,[2,2]]]IEND\0',
                b'ISTART[103,103,[103,'
                b'["Cord 2",10,0,126,1,0,59,0,5,16,1,0,0,"00"]'
                b']]IEND\0',
            ],
            id='detect_door',
        ),
        pytest.param(
            G90SensorUserFlags.DOOR_CHIME, True, False, [
                b'ISTART[102,102,[102,[1,10]]]IEND\0',
                b'ISTART[102,102,[102,[2,2]]]IEND\0',
                b'ISTART[103,103,[103,'
                b'["Cord 2",10,0,126,1,0,55,0,5,16,1,0,0,"00"]'
                b']]IEND\0',
            ],
            id='door_chime',
        ),
        pytest.param(
            G90SensorUserFlags.INDEPENDENT_ZONE, True, False, [
                b'ISTART[102,102,[102,[1,10]]]IEND\0',
                b'ISTART[102,102,[102,[2,2]]]IEND\0',
                b'ISTART[103,103,[103,'
                b'["Cord 2",10,0,126,1,0,47,0,5,16,1,0,0,"00"]'
                b']]IEND\0',
            ],
            id='independent_zone',
        ),
    ]
)
async def test_sensor_user_flag(
    flag: G90SensorUserFlags,
    value_before: bool,
    value_after: bool,
    expected_data: list[bytes], mock_device: DeviceMock
) -> None:
    """
    Tests for settings flags on a sensor.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    sensors = await g90.sensors
    assert sensors[1].get_flag(flag) == value_before
    await sensors[1].set_flag(flag, value_after)
    assert sensors[1].get_flag(flag) == value_after
    assert await mock_device.recv_data == expected_data


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[2,1,2],'
    b'["Cord 1",11,0,126,1,0,63,0,5,16,1,0,""],'
    b'["Cord 2",10,0,126,1,0,63,0,5,16,1,0,""]'
    b']]IEND\0',
    b'ISTART[102,'
    b'[[2,2,1],'
    b'["Cord 2",10,0,126,1,0,63,0,5,16,1,0,""]'
    b']]IEND\0',
    b"ISTARTIEND\0",
])
@pytest.mark.parametrize(
    'value_before,value_after,expected_response', [
        pytest.param(
            G90SensorAlertModes.ALERT_WHEN_AWAY_AND_HOME,
            G90SensorAlertModes.ALERT_ALWAYS, [
                b'ISTART[102,102,[102,[1,10]]]IEND\0',
                b'ISTART[102,102,[102,[2,2]]]IEND\0',
                b'ISTART[103,103,[103,'
                b'["Cord 2",10,0,126,1,0,31,0,5,16,1,0,0,"00"]'
                b']]IEND\0',
            ],
            id='always',
        ),
        pytest.param(
            G90SensorAlertModes.ALERT_WHEN_AWAY_AND_HOME,
            G90SensorAlertModes.ALERT_WHEN_AWAY, [
                b'ISTART[102,102,[102,[1,10]]]IEND\0',
                b'ISTART[102,102,[102,[2,2]]]IEND\0',
                b'ISTART[103,103,[103,'
                b'["Cord 2",10,0,126,1,0,95,0,5,16,1,0,0,"00"]'
                b']]IEND\0',
            ],
            id='when-away',
        ),
        pytest.param(
            G90SensorAlertModes.ALERT_WHEN_AWAY_AND_HOME,
            G90SensorAlertModes.ALERT_WHEN_AWAY_AND_HOME, [
                b'ISTART[102,102,[102,[1,10]]]IEND\0',
            ],
            id='not-changed',
        ),
    ]
)
async def test_sensor_user_alert_mode(
    value_before: G90SensorAlertModes,
    value_after: G90SensorAlertModes,
    expected_response: list[bytes], mock_device: DeviceMock
) -> None:
    """
    Tests for setting alert mode on a sensor.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    sensors = await g90.sensors
    assert sensors[1].alert_mode == value_before
    await sensors[1].set_alert_mode(value_after)
    assert sensors[1].alert_mode == value_after
    assert await mock_device.recv_data == expected_response


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[2,1,2],'
    b'["Cord 1",11,0,126,1,0,33,0,5,16,1,0,""],'
    b'["Cord 2",10,0,126,1,0,33,0,5,16,1,0,""]'
    b']]IEND\0',
    b'ISTART[102,'
    b'[[2,2,1],'
    b'["Cord 2",10,0,126,1,0,1,0,5,16,1,0,""]'
    b']]IEND\0',
    b"ISTARTIEND\0",
])
async def test_sensor_disable_externally_modified(
    mock_device: DeviceMock
) -> None:
    """
    Tests for disabling a sensor that has been modified externally.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    sensors = await g90.sensors
    assert sensors[1].enabled
    await sensors[1].set_enabled(False)
    assert sensors[1].enabled
    assert await mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,10]]]IEND\0',
        b'ISTART[102,102,[102,[2,2]]]IEND\0',
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[1,1,1],["Unsupported",10,0,255,0,0,33,0,0,17,1,0,""]'
    b']]IEND\0',
    b"ISTARTIEND\0",
])
async def test_sensor_unsupported_disable(mock_device: DeviceMock) -> None:
    """
    Tests for disabling an unsupported sensor.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    sensors = await g90.sensors
    assert sensors[0].enabled
    await sensors[0].set_enabled(False)
    assert await mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,10]]]IEND\0',
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[2,1,2],'
    b'["Cord 1",11,0,126,1,0,33,0,5,16,1,0,""],'
    b'["Cord 2",10,0,126,1,0,33,0,5,16,1,0,""]'
    b']]IEND\0',
    b'ISTART[102,[[2,2,0]]]IEND\0',
])
async def test_sensor_disable_sensor_not_found_on_refresh(
    mock_device: DeviceMock
) -> None:
    """
    Tests for disabling a sensor that is not found on refresh.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    sensors = await g90.sensors
    assert sensors[1].enabled
    await sensors[1].set_enabled(False)
    assert sensors[1].enabled
    assert await mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,10]]]IEND\0',
        b'ISTART[102,102,[102,[2,2]]]IEND\0',
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[1,1,1],'
    b'["Cord 2",10,0,126,1,0,33,0,5,16,1,0,""]'
    b']]IEND\0',
    b'ISTART[102,'
    b'[[1,1,1],'
    b'["Cord 2",10,0,126,1,0,33,0,5,16,1,0,""]'
    b']]IEND\0',
    b"ISTARTIEND\0",
])
@pytest.mark.parametrize(
    'flags,expected_response', [
        pytest.param(
            # Intentionally contains non-user settable flag, which should be
            # ignored and not configured for the sensor that initial doesn't
            # have it set
            G90SensorUserFlags.INDEPENDENT_ZONE | G90SensorUserFlags.ARM_DELAY
            | G90SensorUserFlags.SUPPORTS_UPDATING_SUBTYPE, [
                b'ISTART[102,102,[102,[1,10]]]IEND\0',
                b'ISTART[102,102,[102,[1,1]]]IEND\0',
                b'ISTART[103,103,[103,'
                b'["Cord 2",10,0,126,1,0,18,0,5,16,1,0,0,"00"]'
                b']]IEND\0',
            ],
            id='set-flags',
        ),
        pytest.param(
            G90SensorUserFlags.ALERT_WHEN_AWAY_AND_HOME
            | G90SensorUserFlags.ENABLED, [
                b'ISTART[102,102,[102,[1,10]]]IEND\0',
                b'ISTART[102,102,[102,[1,1]]]IEND\0',
            ],
            id='set-flags-not-changed',
        ),
    ]
)
async def test_sensor_set_user_flags(
    flags: G90SensorUserFlags, expected_response: list[bytes],
    mock_device: DeviceMock
) -> None:
    """
    Tests for setting user flags on a sensor.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    sensors = await g90.sensors
    await sensors[0].set_user_flags(flags)
    assert await mock_device.recv_data == expected_response


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[1,1,1],'
    b'["Cord 2",10,0,126,1,0,33,0,5,16,1,0,""]'
    b']]IEND\0',
    b"ISTARTIEND\0",
])
async def test_sensor_delete(mock_device: DeviceMock) -> None:
    """
    Tests deleting the sensor.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    sensors = await g90.sensors
    await sensors[0].delete()
    assert sensors[0].is_unavailable
    assert await mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,10]]]IEND\0',
        b'ISTART[131,131,[131,[10]]]IEND\0',
    ]


REGISTER_SENSOR_SENT_DATA = [
    b'ISTART[102,'
    b'[[2,1,2],'
    b'["Cord 1",11,0,126,1,0,63,0,5,16,1,0,""],'
    b'["Cord 2",10,0,126,1,0,63,0,5,16,1,0,""]'
    b']]IEND\0',
    b'ISTARTIEND\0',
    b'ISTART[102,'
    b'[[3,1,3],'
    b'["Cord 1",11,0,126,1,0,63,0,5,16,1,0,""],'
    b'["Cord 2",10,0,126,1,0,63,0,5,16,1,0,""],'
    b'["Test sensor",3,0,1,3,0,33,0,0,0,1,0,""]'
    b']]IEND\0',
]


@pytest.mark.parametrize(
    'expected_exception', [
        # Normal operation, no exception expected
        pytest.param(
            nullcontext(),
            id='register-sensor-ok',
            marks=pytest.mark.g90device(
                sent_data=REGISTER_SENSOR_SENT_DATA,
                notification_data=[b'[170,[4,[3, "Test sensor", 1]]]\0']
            ),
        ),
        # Operartion simulates the panel sent the notification for the newly
        # added sensor, but it was not found in the list of sensors then
        pytest.param(
            pytest.raises(G90EntityRegistrationError),
            id='register-sensor-not-found',
            marks=pytest.mark.g90device(
                sent_data=REGISTER_SENSOR_SENT_DATA,
                notification_data=[b'[170,[4,[33, "Test sensor", 1]]]\0']
            ),
        ),
        # Operation simulates the panel did not send the notification
        # for the newly added sensor
        pytest.param(
            pytest.raises(G90EntityRegistrationError),
            id='register-sensor-timed-out',
            marks=pytest.mark.g90device(
                sent_data=REGISTER_SENSOR_SENT_DATA,
                notification_data=[]
            ),
        ),
        # Operation simulates the panel sent the notification for the newly
        # added sensor, but it retrieving sensor list afterwards resulted in
        # an error
        pytest.param(
            pytest.raises(G90Error),
            id='register-sensor-failed-update',
            marks=pytest.mark.g90device(
                sent_data=[
                    b'ISTART[102,'
                    b'[[2,1,2],'
                    b'["Cord 1",11,0,126,1,0,63,0,5,16,1,0,""],'
                    b'["Cord 2",10,0,126,1,0,63,0,5,16,1,0,""]'
                    b']]IEND\0',
                    b'ISTARTIEND\0',
                    b'garbage',
                ],
                notification_data=[b'[170,[4,[3, "Test sensor", 1]]]\0']
            ),
        ),
    ])
async def test_sensor_register(
    expected_exception: AbstractContextManager[Exception],
    mock_device: DeviceMock
) -> None:
    """
    Tests for registering a sensor.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    # Setup the notifications, since those are required to complete the
    # registration process
    await g90.use_local_notifications(
        notifications_local_host=mock_device.notification_host,
        notifications_local_port=mock_device.notification_port
    )
    await g90.listen_notifications()

    # Attempt to register a new sensor in parallel with simulating the
    # notification from the panel, otherwise the registration will timeout
    task = asyncio.create_task(
        g90.register_sensor('Door Sensor: WRDS01', 'Test sensor', timeout=0.2),
    )
    # Simulate some delay for panel to complete the registration and send the
    # notification
    await asyncio.sleep(0.1)
    await mock_device.send_next_notification()

    # Wait for registration to complete
    await asyncio.wait([task], timeout=0.5)

    with expected_exception:
        # Retrieve the sensor just registered
        sensor = task.result()

        # Verify the sensor is registered properly
        assert isinstance(sensor, G90Sensor)
        assert sensor.name == 'Test sensor'
        assert sensor.index == 3

        # Verify the sequence of commands sent to the panel
        assert await mock_device.recv_data == [
            b'ISTART[102,102,[102,[1,10]]]IEND\0',
            b'ISTART[156,156,[156,'
            b'["Test sensor",0,0,1,3,0,33,0,0,0,1,0,0,"00"]'
            b']]IEND\0',
            b'ISTART[102,102,[102,[1,10]]]IEND\0',
        ]

    await g90.close_notifications()


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[1,1,1],'
    b'["Cord 2",10,0,126,1,0,33,0,5,16,1,0,""]'
    b']]IEND\0',
    b'ISTART[102,'
    b'[[1,1,1],'
    b'["Cord 2",10,0,126,1,0,32,0,5,16,1,0,""]'
    b']]IEND\0',
])
async def test_sensor_update(mock_device: DeviceMock) -> None:
    """
    Tests updating the sensor and callback assocaited.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    # Set up the sensor list change callback to be called when the sensor
    # list is updated
    list_change_cb = MagicMock()
    g90.sensor_list_change_callback = list_change_cb
    future = asyncio.get_running_loop().create_future()
    list_change_cb.side_effect = (
        lambda *args: future.set_result(True)
    )

    # Trigger sensor list update
    sensors = await g90.get_sensors()

    # Verify the callback is called with correct parameters, indicating the
    # sensor has been added to the list
    await asyncio.wait([future], timeout=0.1)
    list_change_cb.assert_called_once_with(sensors[0], True)

    # Subsequently retrieving the list of sensors should result in same
    # callback invoked, but with different parameters
    future = asyncio.get_running_loop().create_future()
    await g90.get_sensors()
    await asyncio.wait([future], timeout=0.1)
    list_change_cb.assert_called_with(sensors[0], False)



# Copyright (c) 2025 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Device list.
"""
from typing import AsyncGenerator, Optional
import logging
import asyncio
from .device import G90Device
from .base_list import G90BaseList
from ..const import G90Commands
from ..definitions.devices import G90DeviceDefinitions
from ..entities.sensor import G90SensorUserFlags
from ..exceptions import G90EntityRegistrationError

_LOGGER = logging.getLogger(__name__)


class G90DeviceList(G90BaseList[G90Device]):
    """
    Device list class.
    """
    async def _fetch(self) -> AsyncGenerator[G90Device, None]:
        """
        Fetch the list of devices from the panel.

        :yields: G90Device: Device entity.
        """
        devices = self._parent.paginated_result(
            G90Commands.GETDEVICELIST
        )

        async for device in devices:
            obj = G90Device(
                *device.data, parent=self._parent, subindex=0,
                proto_idx=device.proto_idx
            )

            yield obj

            # Multi-node devices (first node has already been handled
            # above)
            for node in range(1, obj.node_count):
                obj = G90Device(
                    *device.data, parent=self._parent,
                    subindex=node, proto_idx=device.proto_idx
                )
                yield obj

    async def register(
        self, definition_name: str,
        room_id: int, timeout: float, name: Optional[str] = None,
    ) -> G90Device:
        """
        Register the devices (switches) to the panel.

        Contrary to registering the sensors, the registration of devices does
        not have an associated notification from the panel, hence the list of
        devices is polled to determine when new device is added.

        :param definition_name: Name of the device definition to register.
        :param room_id: ID of the room to assign the device to.
        :param timeout: Timeout in seconds to wait for the device to be added.
        :param name: Optional name for the device, if not provided, the
         name from the definition will be used.
        :raises G90EntityRegistrationError: If the device could not be
         registered or found after the registration.
        :return: G90Device: The registered device entity.
        """
        device_definition = G90DeviceDefinitions.get_by_name(definition_name)
        dev_name = name or device_definition.name

        # Register the device with the panel
        await self._parent.command(
            G90Commands.ADDDEVICE, [
                dev_name,
                # Registering device requires to provide a free index from
                # panel point of view
                await self.find_free_idx(),
                room_id,
                device_definition.type,
                device_definition.subtype,
                device_definition.timeout,
                # Newly registered devices are enabled by default
                G90SensorUserFlags.ENABLED,
                device_definition.baudrate,
                device_definition.protocol,
                device_definition.reserved_data,
                device_definition.node_count,
                device_definition.rx,
                device_definition.tx,
                device_definition.private_data
            ]
        )

        # Confirm the registration of the device to the panel
        res = await self._parent.command(
            G90Commands.SENDREGDEVICERESULT,
            # 1 = register, 0 = cancel
            [1]
        )

        # The command above returns the index of the added device in the
        # device list from panel point of view
        try:
            added_at = next(iter(res))
            _LOGGER.debug('Device added at index=%s', added_at)
        except StopIteration:
            msg = (
                f"Failed to register device '{dev_name}' - response does not"
                ' contain the index in the device list'
            )
            _LOGGER.debug(msg)
            # pylint: disable=raise-missing-from
            raise G90EntityRegistrationError(msg)

        # Update the list of devices polling for the new entity
        # to appear in the list - it takes some time for the panel
        # to process the registration and add the device to the list
        found = None
        for _ in range(int(timeout)):
            # Update the list of devices from the panel
            await self.update()
            # Try to find the device by the index it was added at
            if found := await self.find_by_idx(
                added_at, exclude_unavailable=False
            ):
                break
            await asyncio.sleep(1)

        if found:
            return found

        msg = (
            f"Failed to find the added device '{dev_name}'"
            f' at index {added_at}'
        )
        _LOGGER.debug(msg)
        raise G90EntityRegistrationError(msg)



# Copyright (c) 2025 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Cloud message implementations for G90 alarm systems.

This module provides concrete message classes for cloud communication with G90
alarm systems, including ping messages, discovery messages, status change
notifications, and alarm notifications.
"""
from typing import List, Type, cast, ClassVar, Any, TypeVar
import logging
from dataclasses import dataclass
from datetime import datetime, timezone

from pyg90alarm.cloud.protocol import G90CloudMessageContext

from .protocol import (
    G90CloudMessage, G90CloudStatusChangeReqMessageBase, G90CloudHeader
)
from .const import G90CloudDirection, G90CloudCommand
from ..const import (
    G90AlertStateChangeTypes, REMOTE_CLOUD_HOST, REMOTE_CLOUD_PORT,
    G90AlertTypes, G90AlertSources, G90AlertStates,
)
from ..definitions.base import G90PeripheralTypes
from ..notifications.base import G90DeviceAlert


_LOGGER = logging.getLogger(__name__)
CLOUD_MESSAGE_CLASSES: List[Type[Any]] = []
CloudMessageT = TypeVar('CloudMessageT', bound='G90CloudMessage')


def cloud_message(obj: Type[CloudMessageT]) -> Type[CloudMessageT]:
    """
    Register a cloud message class.

    This decorator registers the cloud message class in the global registry
    and ensures there are no duplicate registrations for the same command/
    source/destination combination.

    :param obj: The cloud message class to register
    :return: The registered cloud message class
    :raises ValueError: If a class with the same command/source/destination is
     already registered
    """
    for cls in CLOUD_MESSAGE_CLASSES:
        if cls.matches(obj):
            # pylint:disable=protected-access
            raise ValueError(
                f"Duplicate command={obj._command}"
                f"/source={obj._source}/destination={obj._destination}"
                f" in {cls} and {obj}"
            )
    CLOUD_MESSAGE_CLASSES.append(obj)
    return obj


@dataclass
class G90CloudPingRespMessage(G90CloudMessage):
    """
    Response message for ping requests.

    A message sent in response to a ping request from the device.
    """
    _format = ''
    _command = G90CloudCommand.HELLO
    _source = G90CloudDirection.DEVICE
    _destination = G90CloudDirection.UNSPECIFIED
    _header_kls = G90CloudHeader


@cloud_message
@dataclass
class G90CloudPingReqMessage(G90CloudMessage):
    """
    Ping request message sent by the device to the cloud server.

    This message is sent every minute as a keepalive mechanism.

    :attr _responses: The possible response message classes
    """
    _format = ''
    _command = G90CloudCommand.HELLO
    _source = G90CloudDirection.DEVICE
    _destination = G90CloudDirection.UNSPECIFIED
    _responses = [G90CloudPingRespMessage]
    _header_kls = G90CloudHeader


@dataclass
class G90CloudHelloAckMessage(G90CloudMessage):
    """
    Acknowledgement message sent by the cloud server in response to a hello
    message.

    This message confirms receipt of the hello request from the device.
    """
    _format = '<B'
    _command = G90CloudCommand.HELLO_ACK
    _source = G90CloudDirection.CLOUD
    _destination = G90CloudDirection.DEVICE

    flag: int = 1


@dataclass
class G90CloudHelloRespMessage(G90CloudMessage):
    """
    Response message from the cloud server to a hello request from a device.
    """
    _format = '<B'
    _command = G90CloudCommand.HELLO
    _source = G90CloudDirection.CLOUD
    _destination = G90CloudDirection.DEVICE

    flag: int = 0x1f


@dataclass
class G90CloudHelloInfoRespMessage(G90CloudMessage):
    """
    Information response message from the cloud server to a device's hello
    request.

    This message contains information about the port the device should use for
    communication.
    """
    _format = '<i'
    _command = G90CloudCommand.HELLO_INFO
    _source = G90CloudDirection.CLOUD
    _destination = G90CloudDirection.DEVICE

    # Actual port is set from context in `__post_init__()` method below
    port: int = 0

    def __post_init__(self, context: G90CloudMessageContext) -> None:
        super().__post_init__(context)
        self.port = context.local_port


@cloud_message
@dataclass
# pylint:disable=too-many-instance-attributes
class G90CloudHelloReqMessage(G90CloudMessage):
    """
    Hello request message sent by the device to the cloud server.

    This message is sent every minute as part of the device's heartbeat
    mechanism.
    """
    _format = '<15sx4i3sx6i'
    _command = G90CloudCommand.HELLO
    _source = G90CloudDirection.DEVICE
    _destination = G90CloudDirection.CLOUD
    _responses = [
        G90CloudHelloAckMessage, G90CloudHelloRespMessage,
        G90CloudHelloInfoRespMessage
    ]

    _guid: bytes
    flag1: int      # Typically is 1
    flag2: int      # Typically is 0
    flag3: int      # Typically is 2
    flag4: int      # Typically is 28672 (0x7000)
    fw_ver: str
    flag5: int      # Typically is 0x2000NNNN
    flag6: int      # Typically is 48 (0x30)
    flag7: int      # Typically is 0
    flag8: int      # Typically is 7
    flag9: int      # Typically is 30 (0x1E)
    flag10: int     # Typically is 30 (0x1E)

    @property
    def guid(self) -> str:
        """
        Get the device GUID as a string.

        :return: The device's GUID decoded from bytes to string
        """
        return self._guid.decode()


@dataclass
class G90CloudHelloDiscoveryRespMessage(G90CloudMessage):
    """
    Discovery response message from the cloud to the device.

    This message contains information about the cloud server's IP, port, and
    timestamp.
    """
    _format = '<16s4i'
    _command = G90CloudCommand.HELLO
    _source = G90CloudDirection.CLOUD_DISCOVERY
    _destination = G90CloudDirection.DEVICE

    # Simulated cloud response always contains known IP address of the vendor's
    # cloud service - that is, all interactions between alarm panel and
    # simulated cloud service will use same IP address for unification (i.e.
    # traffic redicrection will always be used to divert panel's cloud traffic
    # to the simulated cloud service)
    ip_addr: bytes = REMOTE_CLOUD_HOST.encode()
    flag2: int = 0
    flag3: int = 0
    port: int = REMOTE_CLOUD_PORT
    _timestamp: int = 0  # unix timestamp

    def __post_init__(self, context: G90CloudMessageContext) -> None:
        super().__post_init__(context)

        self._timestamp = int(datetime.now(timezone.utc).timestamp())
        _LOGGER.debug(
            "%s: Timestamp added: %s", type(self).__name__, str(self)
        )
        self.ip_addr = context.cloud_host.encode()
        self.port = context.cloud_port

    @property
    def timestamp(self) -> datetime:
        """
        Get the timestamp as a datetime object.

        :return: The message timestamp converted to a datetime object with UTC
         timezone
        """
        return datetime.fromtimestamp(
            self._timestamp, tz=timezone.utc
        )

    def __str__(self) -> str:
        return (
            f"{type(self).__name__}"
            f"({super().__str__()}"
            f", timestamp={self.timestamp})"
        )


@cloud_message
@dataclass
# pylint:disable=too-many-instance-attributes
class G90CloudHelloDiscoveryReqMessage(G90CloudMessage):
    """
    Hello discovery request message sent by the device.

    This message is used during the device discovery process to locate the
    cloud server.
    """
    _format = '<15sx4i3sx3i'
    _command = G90CloudCommand.HELLO
    _source = G90CloudDirection.DEVICE_DISCOVERY
    _destination = G90CloudDirection.CLOUD
    _responses = [G90CloudHelloDiscoveryRespMessage]

    _guid: bytes
    flag1: int      # Typically is 0
    flag2: int      # Typically is 0
    flag3: int      # Typically is 1
    flag4: int      # Typically is 28672 (0x7000)
    fw_ver: str
    flag5: int      # Typically is 0x05050505
    flag6: int      # Typically is 0x06060030
    flag7: int      # Typically is 0x07070707

    @property
    def guid(self) -> str:
        """
        Get the device GUID as a string.

        :return: The device's GUID decoded from bytes to string
        """
        return self._guid.decode()


@cloud_message
@dataclass
# pylint:disable=too-many-instance-attributes
class G90CloudStatusChangeReqMessage(G90CloudStatusChangeReqMessageBase):
    """
    Status change request message from the device to the cloud.

    This message is sent when the device's status changes, such as arming or
    disarming.
    """
    # 68x are typically zeros, while 34x is some garbage from previous
    # notification message (0x22) with its head overwritten with old/new status
    # values
    _format = '<2B34xi68x'
    _command = G90CloudCommand.STATUS_CHANGE
    _source = G90CloudDirection.DEVICE
    _destination = G90CloudDirection.CLOUD
    _type = G90AlertTypes.STATE_CHANGE

    type: int
    _state: G90AlertStateChangeTypes
    _timestamp: int  # Unix timestamp

    @property
    def state(self) -> G90AlertStateChangeTypes:
        """
        Get the state change type.

        :return: The alert state change type
        """
        return G90AlertStateChangeTypes(self._state)

    @property
    def as_device_alert(self) -> G90DeviceAlert:
        """
        Convert the message to a device alert object.

        :return: A G90DeviceAlert object constructed from the message
         properties
        """
        return G90DeviceAlert(
            device_id=self._context.device_id or '',
            state=self.state,
            event_id=self.state,
            zone_name='',
            type=self._type,
            source=G90AlertSources.DEVICE,
            unix_time=self._timestamp,
            resv4=0,
            other='',
        )

    def __str__(self) -> str:
        return (
            f"{type(self).__name__}"
            f"({super().__str__()}"
            f", type={self.type}"
            f", state={repr(self.state)}"
            f", timestamp={self.timestamp})"
        )


@cloud_message
@dataclass
# pylint:disable=too-many-instance-attributes
class G90CloudStatusChangeSensorReqMessage(G90CloudStatusChangeReqMessageBase):
    """
    Status change sensor request message from the device to the cloud.

    This message is sent when a sensor's status changes, such as when motion is
    detected or a door/window is opened.
    """
    _format = '<4B32si68x'
    _command = G90CloudCommand.STATUS_CHANGE
    _source = G90CloudDirection.DEVICE
    _destination = G90CloudDirection.CLOUD
    _type: ClassVar[G90AlertTypes] = G90AlertTypes.SENSOR_ACTIVITY

    type: int
    sensor_id: int
    _sensor_type: G90PeripheralTypes
    _sensor_state: int
    _sensor: bytes
    _timestamp: int  # Unix timestamp

    @property
    def sensor_type(self) -> G90PeripheralTypes:
        """
        Get the sensor type.

        :return: The type of the sensor that triggered the event
        """
        return G90PeripheralTypes(self._sensor_type)

    @property
    def sensor_state(self) -> G90AlertStates:
        """
        Get the sensor state.

        :return: The state of the sensor that triggered the event
        """
        return G90AlertStates(self._sensor_state)

    @property
    def sensor(self) -> str:
        """
        Get the sensor name as a string.

        :return: The sensor name decoded from bytes with null characters
         removed
        """
        return self._sensor.decode().rstrip('\x00')

    @property
    def as_device_alert(self) -> G90DeviceAlert:
        """
        Convert the message to a device alert object.

        :return: A G90DeviceAlert object constructed from the message
         properties
        """
        return G90DeviceAlert(
            device_id=self._context.device_id or '',
            state=self.sensor_state,
            event_id=cast(G90AlertStateChangeTypes, self.sensor_id),
            zone_name=self.sensor,
            type=self._type,
            source=G90AlertSources.SENSOR,
            unix_time=self._timestamp,
            resv4=0,
            other='',
        )

    def __str__(self) -> str:
        return (
            f"{type(self).__name__}"
            f"({super().__str__()}"
            f", type={self.type}"
            f", sensor={repr(self.sensor)}"
            f", sensor id ={self.sensor_id}"
            f", sensor type={repr(self.sensor_type)}"
            f", sensor state={repr(self.sensor_state)}"
            f", timestamp={self.timestamp})"
        )


@cloud_message
@dataclass
# pylint:disable=too-many-instance-attributes
class G90CloudStatusChangeAlarmReqMessage(G90CloudStatusChangeReqMessageBase):
    """
    Status change alarm request message from the device to the cloud.

    This message is sent when an alarm is triggered on the device, such as when
    an intrusion is detected or when the panic button is pressed.
    """
    _format = '<4B32si68x'
    _command = G90CloudCommand.STATUS_CHANGE
    _source = G90CloudDirection.DEVICE
    _destination = G90CloudDirection.CLOUD
    _type: ClassVar[G90AlertTypes] = G90AlertTypes.ALARM

    type: int
    sensor_id: int
    _sensor_type: G90PeripheralTypes
    _sensor_state: int
    _sensor: bytes
    _timestamp: int  # Unix timestamp

    @property
    def sensor_state(self) -> G90AlertStates:
        """
        Get the sensor state for the alarm event.

        :return: The state of the sensor that triggered the alarm
        """
        return G90AlertStates(self._sensor_state)

    @property
    def sensor(self) -> str:
        """
        Get the sensor name as a string.

        :return: The sensor name decoded from bytes with null characters
         removed
        """
        return self._sensor.decode().rstrip('\x00')

    @property
    def sensor_type(self) -> G90PeripheralTypes:
        """
        Get the sensor type for the alarm event.

        :return: The type of sensor that triggered the alarm
        """
        return G90PeripheralTypes(self._sensor_type)

    @property
    def as_device_alert(self) -> G90DeviceAlert:
        """
        Convert the message to a device alert object.

        :return: A G90DeviceAlert object constructed from the message
         properties
        """
        return G90DeviceAlert(
            device_id=self._context.device_id or '',
            state=self.sensor_state,
            event_id=cast(G90AlertStateChangeTypes, self.sensor_id),
            zone_name=self.sensor,
            type=self._type,
            source=G90AlertSources.DEVICE,
            unix_time=self._timestamp,
            resv4=0,
            other='',
        )

    def __str__(self) -> str:
        return (
            f"{type(self).__name__}"
            f"({super().__str__()}"
            f", type={self.type}"
            f", _type={self._type}"
            f", sensor={repr(self.sensor)}"
            f", sensor id ={self.sensor_id}"
            f", sensor type={repr(self.sensor_type)}"
            f", sensor state={repr(self.sensor_state)}"
            f", timestamp={self.timestamp})"
        )


@cloud_message
@dataclass
class G90CloudNotificationMessage(G90CloudMessage):
    """
    Notification message from the device to the cloud server.

    This message carries notification data from the device that may include
    sensor data or other information.
    """
    _format = ''
    _command = G90CloudCommand.NOTIFICATION
    _source = G90CloudDirection.DEVICE
    _destination = G90CloudDirection.CLOUD

    @property
    def as_notification_message(self) -> bytes:
        """
        Extract the notification message payload.

        :return: The raw notification message bytes extracted from the header
         payload
        """
        return self.header.payload[self.size():]

    def __str__(self) -> str:
        return (
            f"{type(self).__name__}"
            f"({super().__str__()}"
            f", notification_message={self.as_notification_message.decode()})"
        )


@cloud_message
@dataclass
class G90CloudCmdRespMessage(G90CloudMessage):
    """
    Command response message sent by the device.

    This message contains command and sequence information.
    """
    _format = '<HiHi2H'
    _command = G90CloudCommand.HELLO
    _source = G90CloudDirection.UNSPECIFIED
    _destination = G90CloudDirection.CLOUD
    _header_kls = G90CloudHeader

    flag1: int
    seq_num1: int
    flag3: int
    seq_num2: int
    cmd: int
    subcmd: int

    @property
    def body(self) -> bytes:
        """
        Extract the response body payload.

        :return: The raw response body bytes extracted from the header payload
        """
        return self.header.payload[self.size():]

    def __str__(self) -> str:
        return (
            f"{type(self).__name__}"
            f"({super().__str__()}"
            f", cmd={self.cmd}"
            f", subcmd={self.subcmd}"
            f", seq_num={self.seq_num1}"
            f", body={self.body.decode()})"
        )



# Copyright (c) 2025 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Implementation of G90 cloud protocol notifications service.

Provides a server that listens for connections from G90 alarm devices and
handles cloud protocol notifications.
"""
from typing import Optional, cast, Callable
import logging
import asyncio
from asyncio.transports import BaseTransport, Transport
from asyncio.protocols import Protocol, BaseProtocol
from asyncio import Future

from .protocol import (
    G90CloudHeader, G90CloudError, G90CloudMessageNoMatch,
    G90CloudMessageContext,
)
from .messages import (
    CLOUD_MESSAGE_CLASSES, G90CloudNotificationMessage,
    G90CloudStatusChangeAlarmReqMessage,
    G90CloudStatusChangeSensorReqMessage,
    G90CloudStatusChangeReqMessage,
    G90CloudHelloReqMessage,
    G90CloudHelloDiscoveryReqMessage,
)
from ..notifications.base import G90NotificationsBase
from ..notifications.protocol import G90NotificationProtocol
from ..const import (REMOTE_CLOUD_HOST, REMOTE_CLOUD_PORT)


_LOGGER = logging.getLogger(__name__)


# pylint:disable=too-many-instance-attributes
class G90CloudNotifications(G90NotificationsBase, asyncio.Protocol):
    """
    Cloud notifications service for G90 alarm systems.

    Implements a server that listens for connections from G90 alarm devices
    and processes cloud protocol messages, with optional forwarding to an
    upstream server.

    :param protocol_factory: Factory function to create notification
     protocol handlers
    :param local_host: Local host to bind the server to
    :param local_port: Local port to bind the server to
    :param upstream_host: Optional upstream host to forward messages to
    :param upstream_port: Optional upstream port to forward messages to
    :param keep_single_connection: Whether to keep only a single device
     connection
    """
    # pylint:disable=too-many-positional-arguments,too-many-arguments
    # pylint:disable=too-many-statements
    def __init__(
        self,
        protocol_factory: Callable[[], G90NotificationProtocol],
        local_host: str, local_port: int,
        upstream_host: Optional[str] = None,
        upstream_port: Optional[int] = None,
        keep_single_connection: bool = True,
    ) -> None:
        super().__init__(protocol_factory)
        self._transport: Optional[Transport] = None
        self._server: Optional[asyncio.Server] = None
        self._local_host = local_host
        self._local_port = local_port
        self._upstream_host = upstream_host
        self._upstream_port = upstream_port
        self._keep_single_connection = keep_single_connection
        self._upstream_transport: Optional[Transport] = None
        self._done_sending: Optional[Future[bool]] = None
        self._upstream_task: Optional[asyncio.Task[None]] = None

    def connection_made(self, transport: BaseTransport) -> None:
        """
        Handle a new connection from a device.

        :param transport: The transport for the new connection
        """
        host, port = transport.get_extra_info('peername')
        _LOGGER.debug('Connection from device %s:%s', host, port)
        if self._keep_single_connection and self._transport:
            _LOGGER.debug(
                'Closing connection previously opened from %s:%s',
                *self._transport.get_extra_info('peername')
            )
            self._transport.close()
        self._transport = cast(Transport, transport)

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """
        Handle connection loss from a device.

        :param exc: Exception that caused the connection loss, if any
        """
        if exc:
            _LOGGER.debug('Device connection error: %s', exc)

        # Mark device ID as unknown when connection with alarm panel is lost
        self.clear_device_id()

        if self._transport:
            self._transport.close()
        self._transport = None

    # pylint:disable=too-many-branches
    def data_received(self, data: bytes) -> None:
        """
        Process data received from a device.

        Parses messages from the device, handles them, and sends appropriate
        responses back to the device simulating the cloud server, unless
        upstream forwarding is configured - the data is passed thru to the
        upstream server unmodified.

        :param data: Bytes received from the device
        """
        if self._transport is None:
            return

        self.set_last_device_packet_time()

        # If upstream connection is configured, pass thru all data from the
        # panel. This is done in a separate task to avoid blocking the main
        # processing - the passhtru mode is somewhat supplementary, hence the
        # task is create-and-forget with no retries, error handling nor
        # ordering guarantees, it is assumed the cloud service will handle
        # those correctly.
        if (
            self._upstream_host is not None
            and self._upstream_port is not None
        ):
            self._upstream_task = asyncio.create_task(self.send_upstream(data))

        host, port = self._transport.get_extra_info('peername')
        _LOGGER.debug(
            'Data received from device %s:%s: %s', host, port, data.hex(' ')
        )

        try:
            while len(data):
                # Instantiate a context for the messages
                ctx = G90CloudMessageContext(
                    device_id=self.device_id,
                    local_host=self._local_host,
                    local_port=self._local_port,
                    cloud_host=REMOTE_CLOUD_HOST,
                    cloud_port=REMOTE_CLOUD_PORT,
                    upstream_host=self._upstream_host,
                    upstream_port=self._upstream_port,
                    remote_host=host,
                    remote_port=port
                )
                found = False

                for cls in CLOUD_MESSAGE_CLASSES:
                    try:
                        msg = cls.from_wire(data, context=ctx)
                    except G90CloudMessageNoMatch:
                        continue

                    _LOGGER.debug("Cloud message received: %s", msg)

                    # Only these messages carry on the device ID, store it so
                    # vefirication will be performed by `_handle_alert()`
                    # method
                    if cls in [
                        G90CloudHelloReqMessage,
                        G90CloudHelloDiscoveryReqMessage,
                    ]:
                        self.device_id = msg.guid

                    if cls == G90CloudNotificationMessage:
                        self.handle(msg.as_notification_message)

                    if cls in [
                        G90CloudStatusChangeAlarmReqMessage,
                        G90CloudStatusChangeSensorReqMessage,
                        G90CloudStatusChangeReqMessage,
                    ]:
                        alert = msg.as_device_alert
                        if alert:
                            self.handle_alert(
                                alert,
                                # Verifying device ID only makes sense when
                                # connection isn't closed each time a message
                                # is received
                                verify_device_id=(
                                    not self._keep_single_connection
                                )
                            )

                    found = True

                    # Sending message to upstream cloud service isn't
                    # configured, simulate the response and send those back to
                    # the panel
                    if not self._upstream_task:
                        self._done_sending = asyncio.Future()
                        for resp in msg.wire_responses(context=ctx):
                            self._transport.write(resp)
                            _LOGGER.debug(
                                'Sending response to device %s:%s: %s',
                                host, port, resp.hex(' ')
                            )
                        # Signal the future that sending is done
                        self._done_sending.set_result(True)

                    hdr = msg.header
                    break

                if not found:
                    _LOGGER.debug(
                        "Unknown command from device, wire data: '%s'",
                        data.hex(' ')
                    )
                    hdr = G90CloudHeader.from_wire(data, context=ctx)

                # Advance to the next message
                data = data[hdr.message_length:]
        except G90CloudError as exc:
            _LOGGER.error('Error processing data from device: %s', exc)

    def upstream_connection_made(self, transport: BaseTransport) -> None:
        """
        Handle successful connection to the upstream server.

        :param transport: The transport for the upstream connection
        """
        self._upstream_transport = cast(Transport, transport)

    def upstream_connection_lost(self, exc: Optional[Exception]) -> None:
        """
        Handle connection loss to the upstream server.

        :param exc: Exception that caused the connection loss, if any
        """
        if exc:
            _LOGGER.debug('Upstream connection error: %s', exc)

        if self._upstream_transport:
            self._upstream_transport.close()
        self._upstream_transport = None

    def upstream_data_received(self, data: bytes) -> None:
        """
        Process data received from the upstream server.

        Forwards the data to the connected device.

        :param data: Bytes received from the upstream server
        """
        self.set_last_upstream_packet_time()

        _LOGGER.debug('Data received from upstream: %s', data.hex(' '))

        if self._transport:
            host, port = self._transport.get_extra_info('peername')
            _LOGGER.debug(
                'Sending upstream data to device %s:%s', host, port
            )
            self._transport.write(data)

    def get_upstream_protocol(self) -> BaseProtocol:
        """
        Create and return a protocol for the upstream connection.

        :return: Protocol for handling the upstream connection
        """
        class UpstreamProtocol(Protocol):
            """
            Protocol for handling the upstream connection.
            """
            def __init__(self, parent: G90CloudNotifications) -> None:
                """
                Initialize the upstream protocol.

                :param parent: The parent notifications service
                """
                self._parent = parent

            def connection_made(self, transport: BaseTransport) -> None:
                """
                Handle successful connection to the upstream server.

                :param transport: The transport for the upstream connection
                """
                self._parent.upstream_connection_made(transport)

            def connection_lost(self, exc: Optional[Exception]) -> None:
                """
                Handle connection loss to the upstream server.

                :param exc: Exception that caused the connection loss, if any
                """
                self._parent.upstream_connection_lost(exc)

            def data_received(self, data: bytes) -> None:
                """
                Process data received from the upstream server.

                :param data: Bytes received from the upstream server
                """
                self._parent.upstream_data_received(data)

        return UpstreamProtocol(self)

    async def send_upstream(self, data: bytes) -> None:
        """
        Send data to the upstream server.

        Creates a connection to the upstream server if one doesn't exist,
        then sends the provided data.

        :param data: Bytes to send to the upstream server
        """
        if not self._upstream_host or not self._upstream_port:
            return

        try:
            if not self._upstream_transport:
                _LOGGER.debug(
                    'Creating upstream connection to %s:%s',
                    self._upstream_host, self._upstream_port
                )
                loop = asyncio.get_running_loop()

                self._upstream_transport, _ = await loop.create_connection(
                    self.get_upstream_protocol,
                    host=self._upstream_host, port=self._upstream_port
                )

            if self._upstream_transport:
                self._upstream_transport.write(data)
                _LOGGER.debug(
                    'Data sent to upstream %s:%s',
                    self._upstream_host, self._upstream_port
                )
        except (OSError, asyncio.TimeoutError) as exc:
            _LOGGER.debug(
                'Error sending data to upstream %s:%s: %s',
                self._upstream_host, self._upstream_port, exc
            )

    async def listen(self) -> None:
        """
        Start listening for connections from devices.

        Creates a server bound to the configured local host and port.
        """
        loop = asyncio.get_running_loop()

        _LOGGER.debug('Creating cloud endpoint for %s:%s',
                      self._local_host,
                      self._local_port)
        self._server = await loop.create_server(
            lambda: self,
            self._local_host, self._local_port
        )

    async def close(self) -> None:
        """
        Close the server and any active connections.

        Waits for any pending operations to complete, then closes the upstream
        connection and the local server.
        """
        # Ensure all responses are sent to the panel
        if self._done_sending:
            await asyncio.wait([self._done_sending])

        # Wait for the upstream task to finish if it exists
        if self._upstream_task:
            await asyncio.wait([self._upstream_task])

        if self._upstream_transport:
            _LOGGER.debug(
                'Closing upstream connection to %s:%s',
                self._upstream_host, self._upstream_port
            )
            self._upstream_transport.close()
            self._upstream_transport = None

        if self._server:
            _LOGGER.debug(
                'No longer listening for cloud connections on %s:%s',
                self._local_host, self._local_port
            )
            self._server.close()
            self._server = None



# Copyright (c) 2025 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Sensor list.
"""
from __future__ import annotations
from typing import AsyncGenerator, Optional, TYPE_CHECKING
import logging
import asyncio

from .sensor import G90Sensor
from .base_list import G90BaseList
from ..const import G90Commands
from ..exceptions import G90EntityRegistrationError, G90Error
from ..definitions.sensors import G90SensorDefinitions
from ..entities.sensor import G90SensorUserFlags

if TYPE_CHECKING:
    from ..alarm import G90Alarm

_LOGGER = logging.getLogger(__name__)


class G90SensorList(G90BaseList[G90Sensor]):
    """
    Sensor list class.
    """
    def __init__(self, parent: G90Alarm) -> None:
        super().__init__(parent)
        self._sensor_change_future: Optional[asyncio.Future[G90Sensor]] = None

    async def _fetch(self) -> AsyncGenerator[G90Sensor, None]:
        """
        Fetch the list of sensors from the panel.

        :yields: G90Sensor: Sensor entity.
        """
        entities = self._parent.paginated_result(
            G90Commands.GETSENSORLIST
        )

        async for entity in entities:
            yield G90Sensor(
                *entity.data, parent=self._parent, subindex=0,
                proto_idx=entity.proto_idx
            )

    async def sensor_change_callback(
        self, idx: int, name: str, added: bool
    ) -> None:
        """
        Sensor change callback.

        Should be invoked from corresponding panel's notification handler to
        finish the registration process.

        :param idx: Sensor index.
        :param name: Sensor name.
        :param added: True if the sensor was added, False if removed.
        """
        _LOGGER.debug(
            "Sensor change callback: name='%s', index=%d, added=%s",
            name, idx, added
        )
        # The method depends on the future to be created before it is called
        if not self._sensor_change_future:
            raise G90EntityRegistrationError(
                "Sensor change callback called without a future"
            )

        # Update the sensor list to get the latest data, unfortunately there is
        # no panel command to get just a single sensor
        try:
            await self.update()
        except G90Error as err:
            _LOGGER.debug(
                "Failed to update the sensor list: %s", err
            )
            # Indicate the error to the caller
            self._sensor_change_future.set_exception(err)
            return

        # Attempt to find the sensor just changed by its index
        if not (found := await self.find_by_idx(
            idx, exclude_unavailable=False
        )):
            msg = (
                f"Failed to find the added sensor '{name}'"
                f' at index {idx}'
            )
            _LOGGER.debug(msg)
            # Indicate the error to the caller
            self._sensor_change_future.set_exception(
                G90EntityRegistrationError(msg)
            )
            return

        # Provide the found sensor as the future result so that the caller
        # can use it
        self._sensor_change_future.set_result(found)

    async def register(
        self, definition_name: str,
        room_id: int, timeout: float, name: Optional[str] = None,
    ) -> G90Sensor:
        """
        Registers sensor to the panel.

        :param definition_name: Sensor definition name.
        :param room_id: Room ID to assign the sensor to.
        :param timeout: Timeout for the registration process.
        :param name: Optional name for the sensor, if not provided, the
         definition name will be used.
        :raises G90EntityRegistrationError: If the registration fails.
        :return: G90Sensor: The registered sensor entity.
        """
        sensor_definition = G90SensorDefinitions.get_by_name(definition_name)
        dev_name = name or sensor_definition.name

        # Future is needed to coordinate the registration process with panel
        # notifications (see `sensor_change_callback()` method)
        self._sensor_change_future = asyncio.get_running_loop().create_future()

        # Register the sensor with the panel
        await self._parent.command(
            G90Commands.ADDSENSOR, [
                dev_name,
                # Registering sensor requires to provide a free index from
                # panel point of view
                await self.find_free_idx(),
                room_id,
                sensor_definition.type,
                sensor_definition.subtype,
                sensor_definition.timeout,
                # Newly registered sensors are enabled by default and set to
                # alarm in away and home modes
                G90SensorUserFlags.ENABLED
                | G90SensorUserFlags.ALERT_WHEN_AWAY_AND_HOME,
                sensor_definition.baudrate,
                sensor_definition.protocol,
                sensor_definition.reserved_data,
                sensor_definition.node_count,
                sensor_definition.rx,
                sensor_definition.tx,
                sensor_definition.private_data
            ]
        )

        # Waiting for the registration to finish, where panel will send
        # corresponding notification processed by `sensor_change_callback()`
        # method, which manipulates the future created above.
        done, _ = await asyncio.wait(
            [self._sensor_change_future], timeout=timeout
        )

        if self._sensor_change_future not in done:
            msg = f"Failed to learn the device '{dev_name}', timed out"
            _LOGGER.debug(msg)
            raise G90EntityRegistrationError(msg)

        # Propagate any exception that might have occurred in
        # `sensor_change_callback()` method during the registration process
        self._sensor_change_future.exception()

        # Return the registered sensor entity
        return self._sensor_change_future.result()



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Protocol entity for G90 alarm panel status.
"""
from __future__ import annotations
from typing import Any, Dict
from dataclasses import dataclass, asdict
from ..const import G90ArmDisarmTypes


@dataclass
class G90HostStatus:
    """
    Interprets data fields of GETHOSTSTATUS command.
    """
    host_status_data: int
    host_phone_number: str
    product_name: str
    mcu_hw_version: str
    wifi_hw_version: str

    @property
    def host_status(self) -> G90ArmDisarmTypes:
        """
        Translates host status data to G90ArmDisarmTypes.
        """
        return G90ArmDisarmTypes(self.host_status_data)

    def _asdict(self) -> Dict[str, Any]:
        """
        Returns the host information as dictionary.
        """
        return asdict(self)



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Represents various configuration aspects of the alarm panel.
"""
from __future__ import annotations
from typing import TYPE_CHECKING, Optional
import logging
from dataclasses import dataclass
from enum import IntFlag

from pyg90alarm.exceptions import G90Error
from ..const import G90Commands
if TYPE_CHECKING:
    from ..alarm import G90Alarm


class G90AlertConfigFlags(IntFlag):
    """
    Alert configuration flags, used bitwise
    """
    AC_POWER_FAILURE = 1
    AC_POWER_RECOVER = 2
    ARM_DISARM = 4
    HOST_LOW_VOLTAGE = 8
    SENSOR_LOW_VOLTAGE = 16
    WIFI_AVAILABLE = 32
    WIFI_UNAVAILABLE = 64
    DOOR_OPEN = 128
    DOOR_CLOSE = 256
    SMS_PUSH = 512
    UNKNOWN1 = 2048
    UNKNOWN2 = 8192


_LOGGER = logging.getLogger(__name__)


@dataclass
class G90AlertConfigData:
    """
    Represents alert configuration data as received from the alarm panel.
    """
    flags_data: int

    @property
    def flags(self) -> G90AlertConfigFlags:
        """
        :return: The alert configuration flags
        """
        return G90AlertConfigFlags(self.flags_data)

    @flags.setter
    def flags(self, value: G90AlertConfigFlags) -> None:
        """
        :param value: The alert configuration flags
        """
        self.flags_data = value.value


class G90AlertConfig:
    """
    Represents alert configuration as received from the alarm panel.
    """
    def __init__(self, parent: G90Alarm) -> None:
        self.parent = parent
        self._cached_data: Optional[G90AlertConfigData] = None

    async def _get(self) -> G90AlertConfigData:
        """
        Retrieves the alert configuration flags directly from the device.

        :return: The alerts configured
        """
        _LOGGER.debug('Retrieving alert configuration from the device')
        res = await self.parent.command(G90Commands.GETNOTICEFLAG)
        data = G90AlertConfigData(*res)
        _LOGGER.debug(
            'Alert configuration: %s, flags: %s', data,
            repr(data.flags)
        )

        # Cache the retrieved data for `flags_with_fallback` property
        self._cached_data = data

        return data

    async def set(self, flags: G90AlertConfigFlags) -> None:
        """
        .. deprecated:: 2.3.0

        This method is deprecated and will always raise a RuntimeError.
        Please use :meth:`set_flag` to set individual flags.
        """
        raise RuntimeError(
            'The set() method is deprecated. Please use set_flag() to set'
            ' individual flags instead.'
        )

    async def _set(self, flags: G90AlertConfigFlags) -> None:
        """
        Sets the alert configuration flags on the device.
        """
        _LOGGER.debug('Setting alert configuration to %s', repr(flags))
        await self.parent.command(G90Commands.SETNOTICEFLAG, [flags.value])

    async def get_flag(self, flag: G90AlertConfigFlags) -> bool:
        """
        :param flag: The flag to check
        """
        return flag in await self.flags

    async def set_flag(self, flag: G90AlertConfigFlags, value: bool) -> None:
        """
        Sets the given flag to the desired value.

        Uses read-modify-write approach.

        :param flag: The flag to set
        :param value: The value to set
        """
        # Retrieve current flags
        current_flags = await self.flags
        # Skip updating the flag if it has the desired value
        if (flag in current_flags) == value:
            _LOGGER.debug(
                'Flag %s already set to %s, skipping update',
                repr(flag), value
            )
            return

        # Set or reset corresponding user flag depending on desired value
        if value:
            current_flags |= flag
        else:
            current_flags &= ~flag

        # Set the updated flags
        await self._set(current_flags)

    @property
    async def flags(self) -> G90AlertConfigFlags:
        """
        :return: Symbolic names for corresponding flag bits
        """
        return (await self._get()).flags

    @property
    async def flags_with_fallback(self) -> Optional[G90AlertConfigFlags]:
        """
        :return: Symbolic names for corresponding flag bits, falling back to
         cached data if device communication fails
        """
        result = None

        try:
            result = (await self._get()).flags
        except G90Error as exc:
            _LOGGER.debug(
                'Retrieving alert config flags resulted in error %s',
                repr(exc)
            )
            if self._cached_data is not None:
                _LOGGER.debug(
                    'Falling back to cached alert configuration flags: %s',
                    repr(self._cached_data.flags)
                )
                result = self._cached_data.flags
            else:
                _LOGGER.debug('No cached alert configuration flags available')

        return result




# Copyright (c) 2026 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Base class for loading/saving dataclasses to a device.
"""
from __future__ import annotations
from typing import TYPE_CHECKING, Type, TypeVar, Optional, ClassVar, Any, Dict
import logging
from dataclasses import dataclass, astuple, asdict
from ..const import G90Commands
if TYPE_CHECKING:
    from ..alarm import G90Alarm


_LOGGER = logging.getLogger(__name__)
S = TypeVar('S', bound='DataclassLoadSave')


@dataclass
class DataclassLoadSave:
    """
    Base class for loading/saving dataclasses to a device.

    There are multiple ways to implement the functionality:
     - Encapsulate the dataclass inside another class that handles
       loading/saving and exposes dataclass fields as properties. The latter
       part gets complex as properties need to be asynchronous, as well as
       added dynamically at runtime to improve maintainability.
     - Inherit from this class, which provides `load` and `save` methods on top
       of standard dataclasses. This is believed to be more concise and easier
       to understand.

    Implementing classes must define `LOAD_COMMAND` and `SAVE_COMMAND` class
    variables to specify which commands to use for loading and saving data.

    Example usage:

        @dataclass
        class G90ExampleConfig(DataclassLoadSave):
            LOAD_COMMAND = G90Commands.GETEXAMPLECONFIG
            SAVE_COMMAND = G90Commands.SETEXAMPLECONFIG
            field1: int
            field2: str

        # Loading data
        config = await G90ExampleConfig.load(G90_alarm_instance)
        print(config.field1, config.field2)

        # Modifying and saving data
        config.field1 = 42
        await config.save()
    """
    LOAD_COMMAND: ClassVar[Optional[G90Commands]] = None
    SAVE_COMMAND: ClassVar[Optional[G90Commands]] = None

    def __post_init__(self) -> None:
        """
        Post-initialization processing.
        """
        # Instance variable to hold reference to parent G90Alarm instance,
        # declared here to avoid being part of dataclass fields
        self._parent: Optional[G90Alarm] = None

    async def save(self) -> None:
        """
        Save the current data to the device.
        """
        assert self.SAVE_COMMAND is not None, '`SAVE_COMMAND` must be defined'
        assert self._parent is not None, 'Please call `load()` first'

        _LOGGER.debug('Setting data to the device: %s', str(self))
        await self._parent.command(
            self.SAVE_COMMAND,
            list(astuple(self))
        )

    @classmethod
    async def load(cls: Type[S], parent: G90Alarm) -> S:
        """
        Create an instance with values loaded from the device.

        :return: An instance of the dataclass loaded from the device.
        """
        assert cls.LOAD_COMMAND is not None, '`LOAD_COMMAND` must be defined'
        assert parent is not None, '`parent` must be provided'

        data = await parent.command(cls.LOAD_COMMAND)
        obj = cls(*data)
        _LOGGER.debug('Loaded data: %s', str(obj))

        obj._parent = parent

        return obj

    def _asdict(self) -> Dict[str, Any]:
        """
        Returns the dataclass fields as a dictionary.

        :return: A dictionary representation.
        """
        return asdict(self)

    def __str__(self) -> str:
        """
        Textual representation of the entry.

        `str()` is used instead of `repr()` since dataclass provides `repr()`
        by default, and it would be impractical to require each ancestor to
        disable that.

        :return: A textual representation.
        """
        return super().__repr__() + f'({str(self._asdict())})'



# Copyright (c) 2025 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Base entities for peripheral definitions.
"""
from __future__ import annotations
from dataclasses import dataclass
from itertools import groupby
from enum import IntEnum
from abc import ABC, abstractmethod
import logging
from ..exceptions import G90PeripheralDefinitionNotFound, G90Error

_LOGGER = logging.getLogger(__name__)


class G90PeripheralProtocols(IntEnum):
    """
    Protocol types for the peripherals.
    """
    RF_1527 = 0
    RF_2262 = 1
    RF_PRIVATE = 2
    RF_SLIDER = 3
    CORD = 5
    WIFI = 4
    USB = 6


class G90PeripheralTypes(IntEnum):
    """
    Peripheral types.
    """
    DOOR = 1
    GLASS = 2
    GAS = 3
    SMOKE = 4
    SOS = 5
    VIB = 6
    WATER = 7
    INFRARED = 8
    IN_BEAM = 9
    REMOTE = 10
    RFID = 11
    DOORBELL = 12
    BUTTONID = 13
    WATCH = 14
    FINGER_LOCK = 15
    SUBHOST = 16
    REMOTE_2_4G = 17
    GAS_VALVE = 18
    CORD_SENSOR = 126
    SOCKET = 128
    SIREN = 129
    CURTAIN = 130
    SLIDINGWIN = 131
    AIRCON = 133
    TV = 135
    TV_BOX = 136
    SMART_SWITCH = 137
    NIGHTLIGHT = 138
    SOCKET_2_4G = 140
    SIREN_2_4G = 141
    SWITCH_2_4G = 142
    TOUCH_SWITCH_2_4G = 143
    CURTAIN_2_4G = 144
    IR_2_4G = 145
    CORD_DEV = 254
    UNKNOWN = 255


class G90PeripheralMatchModes(IntEnum):
    """
    Defines compare (match) mode for the peripheral.
    """
    ALL = 0
    ONLY20BITS = 1
    ONLY16BITS = 2


class G90PeripheralRwModes(IntEnum):
    """
    Defines read/write mode for the peripheral.
    """
    READ = 0
    WRITE = 1
    READ_WRITE = 2


@dataclass(frozen=True)
class G90PeripheralDefinition:
    # pylint: disable=too-many-instance-attributes
    """
    Holds peripheral definition data.
    """
    type: G90PeripheralTypes
    subtype: int
    rx: int
    tx: int
    private_data: str
    rw_mode: G90PeripheralRwModes
    match_mode: G90PeripheralMatchModes
    name: str
    protocol: G90PeripheralProtocols
    timeout: int
    baudrate: int
    node_count: int

    @property
    def reserved_data(self) -> int:
        """
        Peripheral's 'reserved_data' field to be written, combined of match
        and RW mode values bitwise.
        """
        return self.match_mode.value << 4 | self.rw_mode.value


def unique_definitions(
    obj: type[G90PeripheralDefinitionsBase]
) -> type[G90PeripheralDefinitionsBase]:
    """
    Decorator to ensure that peripheral definitions are unique by name
    and type/subtype/protocol.

    :param obj: Class to decorate.
    :return: Decorated class with unique definitions.
    :raises G90Error: If definitions are not unique.
    """
    names = groupby(
        sorted(obj.definitions(), key=lambda x: x.name),
        lambda x: x.name
    )
    type_subtype = groupby(
        sorted(
            obj.definitions(),
            key=lambda x: (x.type, x.subtype, x.protocol)
        ),
        lambda x: (x.type, x.subtype, x.protocol)
    )
    non_unique_names = [
        k for _, group in names if len(k := list(group)) > 1
    ]
    non_unique_types = [
        k for _, group in type_subtype if len(k := list(group)) > 1
    ]

    msgs = []
    if non_unique_names:
        msgs.append(
            f"{obj}: Peripheral definitions have non-unique names: \n"
            f"{non_unique_names}"
        )
    if non_unique_types:
        msgs.append(
            f"{obj}: Peripheral definitions have non-unique types: \n"
            f"{non_unique_types}"
        )

    if msgs:
        raise G90Error('.\n'.join(msgs))

    return obj


class G90PeripheralDefinitionsBase(ABC):
    """
    Base class for peripheral definitions.
    """
    @classmethod
    @abstractmethod
    def definitions(cls) -> list[G90PeripheralDefinition]:
        """
        Get all peripheral definitions.

        :return: List of peripheral definitions.
        """

    @classmethod
    def get_by_id(
        cls, id_type: G90PeripheralTypes, id_subtype: int,
        protocol: G90PeripheralProtocols
    ) -> G90PeripheralDefinition:
        """
        Gets peripheral definition by type, subtype and protocol.

        :param id_type: Peripheral type.
        :param id_subtype: Peripheral subtype.
        :param protocol: Peripheral protocol.
        :raises G90PeripheralDefinitionNotFound: If definition not found.
        """
        for definition in cls.definitions():
            if (
                definition.type == id_type
                and definition.subtype == id_subtype
                and definition.protocol == protocol
            ):
                _LOGGER.debug(
                    "Found peripheral definition by"
                    " type %d, subtype %d and protocol %d: %s",
                    id_type, id_subtype, protocol, definition
                )
                return definition

        raise G90PeripheralDefinitionNotFound(
            "Peripheral definition not found"
            f" by type={id_type}, subtype={id_subtype}"
            f" and protocol={protocol}",
        )

    @classmethod
    def get_by_name(
        cls, name: str
    ) -> G90PeripheralDefinition:
        """
        Gets peripheral definition by name.

        :param name: Peripheral name.
        :raises G90PeripheralDefinitionNotFound: If definition not found.
        """
        for definition in cls.definitions():
            if definition.name == name:
                _LOGGER.debug(
                    "Found peripheral definition by name '%s': %s",
                    name, definition
                )
                return definition

        raise G90PeripheralDefinitionNotFound(
            f"Peripheral definition not found by name='{name}'"
        )



"""
Tests for alarm phone numbers retrieval and configuration.
"""
import pytest
from pyg90alarm.alarm import (
    G90Alarm,
)
from pyg90alarm.local.alarm_phones import G90AlarmPhones
from .device_mock import DeviceMock


@pytest.mark.g90device(sent_data=[
    b'ISTART[114,'
    b'["1234", "11111111", "", "", "", "", "", "", "87654321", "12345678"]'
    b']IEND\0',
    b'ISTARTIEND\0'
])
async def test_alarm_phones_config(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving and modifying alarm phones information from
    the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    # Retrieve alarm phones configuration
    phones = await g90.alarm_phones()
    assert isinstance(phones, G90AlarmPhones)

    # Verify retrieved data
    assert phones.panel_password == '1234'
    assert phones.panel_phone_number == '11111111'
    assert phones.phone_number_1 == ''
    assert phones.phone_number_2 == ''
    assert phones.phone_number_3 == ''
    assert phones.phone_number_4 == ''
    assert phones.phone_number_5 == ''
    assert phones.phone_number_6 == ''
    assert phones.sms_push_number_1 == '87654321'
    assert phones.sms_push_number_2 == '12345678'

    # Modify and save alarm phones configuration
    phones.panel_password = '5678'
    await phones.save()

    # Verify data sent to the device
    assert await mock_device.recv_data == [
        b'ISTART[114,114,""]IEND\0',
        b'ISTART[108,108,[108,'
        b'["5678","11111111","","","","","","","87654321","12345678"]'
        b']]IEND\0'
    ]



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Implements paginated command for G90 alarm panel protocol.
"""
from __future__ import annotations
import logging
from typing import Any, cast
from dataclasses import dataclass
from .base_cmd import G90BaseCommand, G90BaseCommandData
from ..exceptions import G90Error
from ..const import G90Commands

_LOGGER = logging.getLogger(__name__)


@dataclass
class G90PaginationFields:
    """
    Represents structure of the pagination fields used by alarm panel.

    :meta private:
    """
    total: int
    start: int
    nelems: int


class G90PaginatedCommand(G90BaseCommand):
    """
    Implements paginated command for alarm panel protocol.
    """
    # pylint: disable=too-many-positional-arguments,too-many-arguments
    def __init__(
        self, host: str, port: int, code: G90Commands, start: int, end: int,
        **kwargs: Any
    ) -> None:
        self._start = start
        self._end = end
        self._expected_nelems = end - start + 1
        self._nelems = 0
        self._total = 0
        super().__init__(host, port, code, [self._start, self._end],
                         **kwargs)

    @property
    def total(self) -> int:
        """
        Total number of records available.
        """
        return self._total

    @property
    def start(self) -> int:
        """
        Index of the first record in the response.
        """
        return self._start

    @property
    def count(self) -> int:
        """
        Number of records in the response.
        """
        return self._nelems

    def _parse(self, data: str) -> None:
        """
        Parses the response from the alarm panel.
        """
        super()._parse(data)
        resp_data: G90BaseCommandData = self._resp.data or []
        try:
            page_data = resp_data.pop(0)
            page_info = G90PaginationFields(*page_data)
        except TypeError as exc:
            raise G90Error(f'Wrong pagination data {page_data} - {str(exc)}'
                           ) from exc
        except IndexError as exc:
            raise G90Error(f"Missing pagination in response '{self._resp}'"
                           ) from exc

        self._total = page_info.total
        self._start = page_info.start
        self._nelems = page_info.nelems

        errors = []
        if self._nelems != len(resp_data):
            qualifier = (
                "Truncated" if self._nelems > len(resp_data) else "Extra"
            )
            errors.append(
                f'{qualifier} data provided in paginated response -'
                f' expected {self._nelems} entities as per response,'
                f' received {len(resp_data)}')

        if self._expected_nelems < len(resp_data):
            errors.append(
                f'Extra data provided in paginated response -'
                f' expected {self._expected_nelems} entities as per request,'
                f' received {len(resp_data)}')

        if errors:
            raise G90Error('. '.join(errors))

        _LOGGER.debug('Paginated command response: '
                      'total records %s, start record %s, record count %s',
                      page_info.total, page_info.start, page_info.nelems)

    async def process(self) -> G90PaginatedCommand:
        """
        Initiates the command processing.
        """
        return cast(G90PaginatedCommand, await super().process())



# Copyright (c) 2025 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Base entity.
"""
from __future__ import annotations
from abc import ABC, abstractmethod
# `Self` has been introduced in Python 3.11, need to use `typing_extensions`
# for earlier versions
try:
    from typing import Self  # type: ignore[attr-defined,unused-ignore]
except ImportError:
    from typing_extensions import Self


class G90BaseEntity(ABC):
    """
    Base entity class.

    Contains minimal set of method for :class:`.G90BaseList` class
    """
    @abstractmethod
    def update(
        self,
        obj: Self  # pylint: disable=used-before-assignment
    ) -> None:
        """
        Update the entity from another one.

        :param obj: Object to update from.
        """

    @property
    @abstractmethod
    def is_unavailable(self) -> bool:
        """
        Check if the entity is unavailable.

        :return: True if the entity is unavailable.
        """

    @is_unavailable.setter
    @abstractmethod
    def is_unavailable(self, value: bool) -> None:
        """
        Set the entity as unavailable.

        :param value: Value to set.
        """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Get the name of the entity.

        :return: Name of the entity.
        """

    @property
    @abstractmethod
    def index(self) -> int:
        """
        Get the index of the entity.

        :return: Index of the entity.
        """

    @property
    @abstractmethod
    def subindex(self) -> int:
        """
        Get the subindex of the entity.

        :return: Subindex of the entity.
        """



"""
Tests for alarm panel configuration retrieval and modification.
"""
import pytest
from pyg90alarm.alarm import (
    G90Alarm,
)
from pyg90alarm.local.host_config import (
    G90HostConfig, G90SpeechLanguage, G90VolumeLevel,
)
from .device_mock import DeviceMock


@pytest.mark.g90device(sent_data=[
    b'ISTART[106,'
    b'[900,0,0,1,2,2,60,2,0,60,2]'
    b']IEND\0',
    b'ISTARTIEND\0'
])
async def test_host_config(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving and modifying alarm panel configuration from
    the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    # Retrieve configuration
    cfg = await g90.host_config()
    assert isinstance(cfg, G90HostConfig)

    # Verify retrieved values
    assert cfg.alarm_siren_duration == 900
    assert cfg.speech_language == G90SpeechLanguage.ENGLISH_MALE
    assert cfg.speech_volume_level == G90VolumeLevel.HIGH
    assert cfg.timezone_offset_m == 60
    assert cfg.ring_volume_level == G90VolumeLevel.HIGH
    assert cfg.key_tone_volume_level == G90VolumeLevel.MUTE
    assert cfg.ring_duration == 60
    assert cfg.alarm_delay == 0
    assert cfg.arm_delay == 0
    assert cfg.backlight_duration == 1
    assert cfg.alarm_volume_level == G90VolumeLevel.HIGH

    # Modify and save configuration
    cfg.alarm_siren_duration = 600
    await cfg.save()

    # Verify data sent to the device
    assert await mock_device.recv_data == [
        b'ISTART[106,106,""]IEND\0',
        b'ISTART[107,107,[107,'
        b'[600,0,0,1,2,2,60,2,0,60,2]'
        b']]IEND\0'
    ]



# Copyright (c) 2025 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Base entity list.
"""
from abc import ABC, abstractmethod
from typing import (
    List, AsyncGenerator, Optional, TypeVar, Generic, cast, TYPE_CHECKING,
    Callable, Coroutine, Union
)
import asyncio
import logging

from ..exceptions import G90Error
from .base_entity import G90BaseEntity
from ..callback import G90Callback

T = TypeVar('T', bound=G90BaseEntity)
ListChangeCallback = Union[
    Callable[[T, bool], None],
    Callable[[T, bool], Coroutine[None, None, None]]
]

if TYPE_CHECKING:
    from ..alarm import G90Alarm
else:
    # Alias G90Alarm to object avoid circular imports
    # (`G90Alarm` -> `G90SensorList` -> `G90BaseList` -> `G90Alarm`)
    G90Alarm = object

_LOGGER = logging.getLogger(__name__)


class G90BaseList(Generic[T], ABC):
    """
    Base entity list class.

    :param parent: Parent alarm panel instance.
    """
    def __init__(self, parent: G90Alarm) -> None:
        self._entities: List[T] = []
        self._lock = asyncio.Lock()
        self._parent = parent
        self._list_change_cb: Optional[ListChangeCallback[T]] = None

    @abstractmethod
    async def _fetch(self) -> AsyncGenerator[T, None]:
        """
        Fetch the list of entities from the panel.

        :return: Async generator of entities
        """
        yield cast(T, None)  # pragma: no cover

    @property
    async def entities(self) -> List[T]:
        """
        Return the list of entities.

        :meth:`update` is called if the list is empty.

        :return: List of entities
        """
        # Please see below for the explanation of the lock usage
        async with self._lock:
            entities = self._entities

        if not entities:
            return await self.update()

        return entities

    async def update(self) -> List[T]:
        """
        Update the list of entities from the panel.

        :return: List of entities
        """
        # Use lock around the operation, to ensure no duplicated entries in the
        # resulting list or redundant exchanges with panel are made when the
        # method is called concurrently
        async with self._lock:
            entities = self._fetch()

            non_existing_entities = self._entities.copy()
            try:
                async for entity in entities:
                    try:
                        existing_entity = next(
                            x for x in self._entities if x == entity
                        )
                    except StopIteration:
                        existing_entity = None

                    if existing_entity is not None:
                        # Update the existing entity with the new data
                        _LOGGER.debug(
                            "Updating existing entity '%s' from protocol"
                            " data '%s'", existing_entity, entity
                        )

                        existing_entity.update(entity)
                        non_existing_entities.remove(existing_entity)

                        # Invoke the list change callback for the existing
                        # entity to notify about the update
                        G90Callback.invoke(
                            self._list_change_cb, existing_entity, False
                        )
                    else:
                        # Add the new entity to the list
                        _LOGGER.debug('Adding new entity: %s', entity)
                        self._entities.append(entity)
                        # Invoke the list change callback for the new entity
                        G90Callback.invoke(self._list_change_cb, entity, True)
            except TypeError as err:
                _LOGGER.error(
                    'Failed to fetch entities: %s', err
                )
                raise G90Error(err) from err

            # Mark the entities that are no longer in the list
            for unavailable_entity in non_existing_entities:
                _LOGGER.debug(
                    'Marking entity as unavailable: %s', unavailable_entity
                )
                unavailable_entity.is_unavailable = True

            _LOGGER.debug(
                'Total number of entities: %s, unavailable: %s',
                len(self._entities), len(non_existing_entities)
            )

            return self._entities

    async def find_by_idx(
        self, idx: int, exclude_unavailable: bool, subindex: int = 0
    ) -> Optional[T]:
        """
        Finds entity by index.

        :param idx: Entity index
        :param exclude_unavailable: Exclude unavailable entities
        :param subindex: Entity subindex
        :return: Entity instance or None if not found
        """
        entities = await self.entities

        found = None
        if idx < len(entities):
            entity = entities[idx]
            if entity.index == idx and entity.subindex == subindex:
                # Fast lookup by direct index
                _LOGGER.debug('Found entity via fast lookup: %s', entity)
                found = entity

        if not found:
            for entity in entities:
                if entity.index == idx and entity.subindex == subindex:
                    _LOGGER.debug('Found entity: %s', entity)
                    found = entity

        if found:
            if not exclude_unavailable or not found.is_unavailable:
                return found

            _LOGGER.debug(
                'Entity is found but unavailable, will result in none returned'
            )

        _LOGGER.error(
            'Entity not found by index=%s and subindex=%s', idx, subindex
        )
        return None

    async def find(
        self, idx: int, name: str, exclude_unavailable: bool, subindex: int = 0
    ) -> Optional[T]:
        """
        Finds entity by index, subindex and name.

        :param idx: Entity index
        :param name: Entity name
        :param exclude_unavailable: Exclude unavailable entities
        :param subindex: Entity subindex
        :return: Entity instance or None if not found
        """
        found = await self.find_by_idx(idx, exclude_unavailable, subindex)
        if not found:
            return None

        if found.name == name:
            return found

        _LOGGER.error(
            'Entity not found: index=%s, subindex=%s, name=%s',
            idx, subindex, name
        )
        return None

    async def find_free_idx(self) -> int:
        """
        Finds the first free index in the list.

        The index is from protocol point of view (`.index` attribute of the
        protocol data), not the index in the list. The index is required when
        registering a new entity on the panel.

        :return: Free index
        """
        entities = await self.entities

        # Collect indexes in use by the existing entities
        occupied_indexes = set(x.index for x in entities)
        # Generate a set of possible indexes from 0 to the maximum index in
        # use, or provide an empty set if there are no existing entities
        if occupied_indexes:
            possible_indexes = set(range(0, max(occupied_indexes)))
        else:
            # No occupied indexes, so possible_indexes is empty
            possible_indexes = set()

        try:
            # Find the first free index by taking difference between
            # possible indexes and occupied ones, and then taking the minimum
            # value off the difference
            free_idx = min(
                possible_indexes.difference(occupied_indexes)
            )
        except ValueError:
            # If no gaps in existing indexes, then return the index next to
            # the last existing entity
            free_idx = len(entities)

        _LOGGER.debug(
            'Found free index=%s out of occupied indexes: %s',
            free_idx, occupied_indexes
        )
        return free_idx

    @property
    def list_change_callback(self) -> Optional[ListChangeCallback[T]]:
        """
        List change callback.

        Invoked when the list of entities is changed, i.e. when a new entity is
        added or an existing one is updated.

        :return: Callback
        """
        return self._list_change_cb

    @list_change_callback.setter
    def list_change_callback(self, value: ListChangeCallback[T]) -> None:
        self._list_change_cb = value
